/*
 * stbadmin module
 *
 * author: Laser
 *
 *
 *
 *
 *
 */
#include "boa.h"
#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <sys/signal.h>
#include <stdarg.h>

extern int isblank(int c);
extern char *nvram_get(const char *name);
extern int nvram_set(const char *name, const char *value);
extern int nvram_commit(void);
static void set_timer(int sec, void (*handler)(int t));


#define ARGV(args...) ((char *[]) { args, NULL })

static int m_gb_upgrade = 0;
#define ASSERT_UPGRADE(mask) {if( m_gb_upgrade && !mask ) return 0;}

typedef struct _NW_List_Item {
	char ssid[32];
	char channel[4];
	char mode[4];
	char security[12];
	char encryption[8];
	char strength[4];

	int sig;
	struct _NW_List_Item *next;
} ListItem;

int echo(char *fmt, ...)
{
	return 0;

	static FILE *fp = NULL;
	if( !fp ) {
		fp = fopen("/dev/console", "w");
		if( !fp )
			return 0;
	}

	va_list ap;
	va_start(ap, fmt);
	int t = vfprintf (fp, fmt, ap);
	va_end(ap);

	return t;
}

inline int
valid_choice(char *value, char *argv[])
{
	char **choice = argv;
	
	for (choice = argv; *choice ; choice++) {
		if (!strcmp(value, *choice))
			return 1;
	}
	
	return 0;
}

inline int
valid_range(char *value, char *argv[])
{
	int n, start, end;
	
	n = atoi(value);
	start = atoi(argv[0]);
	end = atoi(argv[1]);
	
	if (n < start || n > end) {
		return 0;
	}
	
	return 1;
}

inline int
valid_name(char *value, char *argv[])
{
	int n, min, max;
	
	n = strlen(value);
	min = atoi(argv[0]);
	max = atoi(argv[1]);
	
	if (n > max || n < min) {
		return 0;
	}
	
	return 1;
}

void dump_uri(request * req, char *uri, char *query, char *name[], char *value[], char *title)
{
	SQUASH_KA(req);
	req->response_status = R_REQUEST_OK;
	req_write(req, "HTTP/1.0 200 OK\r\n");
	print_http_headers(req);
	req_write(req, "Content-Type: " "text/html; charset=ISO-8859-1" "\r\n\r\n"); /* terminate header */

	char buffer[MAX_HEADER_LENGTH] = {0}, *p = buffer;
	p += sprintf(p, "URI = %s\r\nQuery = %s\r\n\r\n", uri, query);

	int i;
	for(i = 0; name[i] && value[i] ; i++ )
		p += sprintf(p, "name  = %s\r\nvalue = %s\r\n\r\n", name[i], value[i]);

	req_write(req,	"<HTML><HEAD><TITLE>");
	req_write(req, title);
	req_write(req, "</TITLE></HEAD>\n<BODY><H1>");
	req_write(req, title);
	req_write(req, "</H1>\n");
	req_write(req, buffer);
	req_write(req,"</BODY></HTML>\n");
	req_flush(req);
}

void send_error_response(request *req, char *code)
{
	SQUASH_KA(req);
        req->response_status = R_REQUEST_OK;
        req_write(req, "HTTP/1.0 ");
	req_write(req, code);
	req_write(req, " OK\r\n");
	req_write(req, "Content-Type: text/plain\r\n");
	req_write(req, "Content-length: 0\r\n\r\n");
	req_flush(req);
}

void print_header(request *req)
{
        SQUASH_KA(req);
        req->response_status = R_REQUEST_OK;
        req_write(req, "HTTP/1.0 200 OK\r\n");
	req_write(req, "Content-Type: text/plain\r\n");
}

static void wireless_reboot(int i)
{
	system("/usr/sbin/wlconf eth1 down > /dev/console 2>&1");
	system("/usr/sbin/wlconf eth1 up > /dev/console 2>&1");

	// restart the lan
	system("killall udhcpc");
	system("/sbin/lan restart > /dev/console 2>&1");

	nvram_set("dhcp_flg", "1");
	nvram_commit();
}

static int networks_insert_item(ListItem **head, char *ssid, char *channel, char *mode, char *security, char *encryption, char *strength)
{
	ListItem *pre = NULL, *item = NULL, *next = NULL;

	// check the item validation
	if( !strlen(ssid) || !strlen(channel) || !strlen(mode) || !strlen(security) || !strlen(encryption) || !strlen(strength) )
		return 1;

	item = (ListItem *)calloc(1, sizeof(ListItem));
	strcpy(item->ssid, ssid);
	strcpy(item->channel, channel);
	strcpy(item->mode, mode);
	strcpy(item->security, security);
	strcpy(item->encryption, encryption);
	strcpy(item->strength, strength);
	item->sig = strtol(strength, NULL, 10);

	if( !*head ) { // have not any one
		*head = item;
	} else {
		pre = *head;

		if( (pre->sig < item->sig) || (pre->sig == item->sig && strcmp(ssid, pre->ssid) <= 0) ) {
			item->next = pre;
			*head = item;
		} else {
			while( pre ) {
				next = pre->next;
				if( !next ) {
					pre->next = item;
					break;
				} else {
					if( (item->sig > next->sig) || (item->sig == next->sig && strcmp(ssid, next->ssid) <= 0) ) {
						item->next = next;
						pre->next = item;
						break;
					} else {
						pre = pre->next;
					}
				}
			}
		}
	}

	return 0;
}

#define UPGRADE_STATUS "/tmp/stbadmin_upgrade_status" // this file is from broadcom53xx-src/tools/script-for-fw-upgrade/fw-upgrade.sh
static void spy_upgrade_handler(int t)
{
	FILE *fp = fopen(UPGRADE_STATUS, "r");
	if( !fp ) {
		return ;
	}

	char buffer[1024] = {0}, *proc = buffer, *status;
	fgets(buffer, sizeof(buffer), fp);
	fclose(fp);

	if( !(proc = strchr(buffer, ':')) )
		return ;
	*proc = 0;
	status = proc + 2;
	proc = buffer;
	echo("spy_upgrade_handler: proc = %s, status = %s\n", proc, status);

	if( !strcmp(proc, "download") ) { // downloading
	} else if( !strcmp(proc, "upgrade") ) {  // upgrading
		set_timer(0, NULL);
		m_gb_upgrade = 0;
		if( !strncmp(status, "success", 7) ) {
			echo("spy_upgrade_handler: successful to upgrade, rebooting...\n");
			system("kill 1; /sbin/reboot");
		}
		return ;

	} else { // unknown progress
		m_gb_upgrade = 0;
		echo("error: spy_upgrade_handler");
	}
}

/*
 * set a timer
 * param:
 *   wait    -- the timer time, if 0, the timer will be stopped.
 *   handler -- the function when the time elapsed
 */
static void set_timer(int wait, void (*handler)(int t))
{
	if( wait > 0 && handler ) {
		struct sigaction sa;
		sa.sa_handler = handler;
		sigemptyset(&sa.sa_mask);
		sa.sa_flags = SA_RESTART;
		sigaction(SIGALRM, &sa, NULL);
	}

	struct itimerval itv;
	itv.it_value.tv_sec = wait;
	itv.it_value.tv_usec = 0;
	itv.it_interval = itv.it_value;
	
	setitimer(ITIMER_REAL, &itv, 0);
}

/*
 * read_assoc_status -- read the file and perse it
 *
 * return:
 *     0 -- succes
 *     1 -- normal error
 *     2 -- Not associated"
 *
 * example of the file content

# wl assoc
SSID: "NETGEAR"
Mode: Managed   RSSI: -42 dBm   noise: 0 dBm    Channel: 10
BSSID: 00:1F:33:E3:5C:F6        Capability: ESS ShortPre ShortSlot
Supported Rates: [ 1(b) 2(b) 5.5(b) 6 9 11(b) 12 18 24 36 48 54 ]
802.11N Capable:
        Chanspec: 2.4GHz channel 10 20MHz (0x2b0a)
        Control channel: 10
        802.11N Capabilities:
        Supported MCS : [ 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 ]

 */
#define ASSOC_STATUS "/tmp/stbadmin_assoc_status"
#define ASSOC_STATUS_OK     0 // associated
#define ASSOC_STATUS_ERROR  1 // error
#define ASSOC_STATUS_NOT    2 // not associated
#define ASSOC_STATUS_NOCONF 3 // not configured
#define ASSOC_STATUS_NOAUTH 4 // not authenticated
int read_assoc_status(char *ssid, char *security, char *encryption, char *channel, char *mode, char *apmac, char *strength)
{
	system("/usr/sbin/wl assoc > " ASSOC_STATUS);

	FILE *fp = fopen(ASSOC_STATUS, "r");
	if( !fp )
		return 1;

	char tmp[1024 * 2] = {0}, *q;
	int chan = 0;
	int retval = ASSOC_STATUS_OK;

	strcpy(security, "open");
	strcpy(encryption, "none");
	while( !feof(fp) ) {
		memset(tmp, 0, sizeof(tmp));
		q = fgets(tmp, sizeof(tmp), fp);
		if( !q )
			continue;

		// filter
		while( isblank(*q) )
			q++;
		q[strlen(q) - 1] = 0;  // the last \n character

		//echo("data = %s\n", q);
		if( strncmp(q, "Not associated", 14) == 0 ) {
			if( strlen(q) == 45 )
				retval = ASSOC_STATUS_NOCONF;
			else
				retval = ASSOC_STATUS_NOT;
			break;
		}

		if( strncmp("SSID", q, 4) == 0 ) { // SSID
			q = strchr(q, ':');
			q += 3;
			q[strlen(q) - 1] = 0;
			strcpy(ssid, q);

			//echo("ssid = %s\n", ssid);
			continue;
		}

		if( strncmp("Mode", q, 4) == 0 ) { // Mode, RSSI, Channel
			char *s_rssi = strstr(q, "RSSI") + 6;
			int d_rssi = 0;
			sscanf(s_rssi, "%d", &d_rssi);
			sprintf(strength, "%d", (d_rssi = (int)((d_rssi + 95)*100/85)) > 100 ? 100 : d_rssi );

			q = strstr(q, "Channel") + strlen("Channel") + 2;
			strcpy(channel, q);
			sscanf(channel, "%d", &chan);

			//echo("channel = %s, strength = %s\n", channel, strength);
			continue;
		}

		if( strncmp("BSSID", q, 5) == 0 ) {
			memcpy(apmac, q + 7, 17);

			//echo("apmac = %s\n", apmac);
			continue;
		}

		if( strncmp("Supported Rates", q, 15) == 0 ) {
			// the rate format: [ 1(b) 2(b) 5.5(b) 6 9 11(b) 12 18 24 36 48 54 ]
			// b rate: 1, 2, 5.5, 11
			// g rate: 6, 9, 12, 18, 24, 36, 48, 54
			q += 15 + 2;

			int brate = 0, grate = 0, rate;
			char *pr = strtok(q, " ");
			while( pr ) {
				if( sscanf(pr, "%d", &rate) > 0 ) {
					switch( rate ) {
						case 1: case 2: case 5: case 11:
							brate = 1;
							break;
						case 6: case 9: case 12: case 18: case 24: case 36: case 48: case 54:
							grate = 1;
							break;
						default:
							echo("Wrong Rates");
							break;
					}
				}

				pr = strtok(NULL, " ");
			}
			if( brate == 1 && grate == 0 )
				strcpy(mode, "b");
			if( brate == 0 && grate == 1 )
				strcpy(mode, "g");
			if( brate && grate )
				strcpy(mode, "g");

			//echo("mode = %s\n", mode);
			break;
		}
	}
	fclose(fp);
	unlink(ASSOC_STATUS);

	// get the security and encryption
	static char sec_list[][12] = {{"open"}, {"shared"}, {"wpa-psk"}, {"wpa2-psk"}};
	char sec[4] = {0}, key_leng[8] = {0};
	int  index = 0;
	strcpy(sec, nvram_get("wl_sectype") ? : "0");
	strcpy(key_leng, nvram_get("key_length") ? : "64");

	index = sec[0] - '0';
	if( index < 0 || index > 3 ) {
		echo("Wrong Security index: %d\n", index);
		return 0;
	}

	strcpy(security, sec_list[index]);
	if( 0 == index ) {
		strcpy(encryption, "none");
	} else if( 1 == index ) {
		strcpy(encryption, key_leng[0] == '6' ? "wep64" : "wep128");
		strcpy(security, nvram_get("wl0_auth_mode") ? : "open");
		if( !strcmp(security, "none") )
			strcpy(security, "open");
	} else if( 2 == index ) {
		strcpy(encryption, "tkip");
	} else {
		strcpy(encryption, "aes");
	}

	return retval;
}

static void get_firmware_version(char *fw, int leng)
{
	if( m_gb_upgrade == 0 ) {
		FILE *fp = fopen("/etc/firmware.version", "r");
		if( !fp )
			return;
		fgets(fw, leng, fp);
		fclose(fp);

		char *p = fw;
		p += strlen(p) - 1;
		while( p && (*p == '\n' || *p == '\r' ) )
			p--;
		*++p = 0;
	} else { // is upgrading
		FILE *fp = fopen("/tmp/fw_info", "r"); // the file is specified in file "broadcom53xx-src/tools/script/fw-upgrade.sh"
		if( !fp )
			return;
		fgets(fw, leng, fp);
		fclose(fp);

		char *p = fw + strlen(fw) - 1, *q = 0;
		while( p && (*p == '\n' || *p == '\r' || *p == '#' ) )
			p--;
		*++p = 0;

		q = fw, p = fw + 8; // "wg602v4-"
		while( *p )
			*q++ = *p++;
		*q = 0;
	}
}

#define RESOLV_CONF "/tmp/resolv.conf"
static char* get_dns()
{
	static char dns[256] = {0};
	char tmp[64], *p, *q = dns;
	FILE *fp = fopen(RESOLV_CONF, "r");
	if( !fp )
		return dns;

	memset(dns, 0, sizeof(dns));
	while( !feof(fp) ) {
		if( !(p = fgets(tmp, sizeof(tmp), fp)) )
			break;

		// remove the '\n' or '\r' at the end
		while( *p && *p >= 32 )
			p++;
		*p = 0;

		if( !strncmp(tmp, "nameserver", 10) && q + 15 < dns + sizeof(dns)) {
			q += sprintf(q, "%s ", tmp + 11);
		}
	}
	fclose(fp);

	if( strlen(dns) )
		*(--q) = 0;
	return dns;
}

/*
 * check whether a ip address is valid or not
 *
 */
static int check_address_valid(char *str, int mask)
{
	char data[32] = {0}, *token = NULL;
	int subi = 0, field = 0;
	unsigned char fds[4] = {0};
	unsigned int ip = 0;

	strcpy(data, str);
	token = strtok(data, ".");
	while( token && field < 4 ) {
		subi = strtol(token, NULL, 10);
		if( subi < 0 || subi > 255 )
			return 0;
		ip = ip<<8 | (subi&0xff);
		fds[field++] = subi&0xff;
		
		token = strtok(NULL, ".");
	}

	if( field < 4 || token || ip == 0x7f000001 )
		return 0;

	// fds[0]: 1 -223, 224 - 255 is multicast address
	// fds[1]: 0 - 255
	// fds[2]: 0 - 255
	// fds[3]: 1 - 254
	if( !mask && !(fds[0] >= 1 && fds[0] <= 223 && fds[3] >= 1 && fds[3] <= 254) )
		return 0;

	if( mask ) { // check subnet mask
		if( ip == 0 || ip == 0xffffffff )
			return 0;
		
		int b = 0;
		while( ip & 0x80000000 ) {
			ip <<= 1;
			b++;
		}
		if( b < 32 ) {
			b = 32 - b;
			while( b-- ) {
				if( ip & 0x80000000 )
					return 0;
				ip <<= 1;
			}
		}
	}
	
	return 1;
}

#define nvram_safe_get(name) (nvram_get(name) ? : "")
int stbadmin_info(request *req, char *name[], char *value[], int sum)
{
	ASSERT_UPGRADE(1);

	char buffer[MAX_HEADER_LENGTH] = {0}, *p = buffer, fw[32] = {0}, bridge_name[36] = {0}, lan_mac[32] = {0};
	char ssid[36] = {0}, security[12] = {0}, encryption[8] = {0}, channel[4] = {0}, mode[4] = {0}, apmac[24] = {0}, strength[4] = {0};
	char gw[32] = {0}, lan_ip[32] = {0}, lan_mask[32] = {0};
	int  is_def = 0;

	get_firmware_version(fw, sizeof(fw));
	strcpy(bridge_name, nvram_safe_get("netbiosname"));
	strcpy(lan_mac, nvram_safe_get("lan_hwaddr"));
	strcpy(gw, nvram_safe_get("lan_gateway"));
	strcpy(lan_ip, nvram_safe_get("lan_ipaddr"));
	strcpy(lan_mask, nvram_safe_get("lan_netmask"));

	if( strlen(bridge_name) == 0 )
		strcpy(bridge_name, "netgear");
	if( strncmp("0.0.0.0", gw, 7) == 0 )
		memset(gw, 0, sizeof(gw));
	if( !strcmp("192.168.0.227", lan_ip) && !strcmp(lan_mask, "255.255.255.0") )
		is_def = 1; // blank

	p += sprintf(p, "protocol=\"2\"\r\n");
	p += sprintf(p, "manufacturer=\"netgear\"\r\n");
	p += sprintf(p, "model=\"wg602v4\"\r\n");
	p += sprintf(p, "firmware=\"%s\"\r\n", fw);
	p += sprintf(p, "skyid=\"skywirelessbridge\"\r\n");
	p += sprintf(p, "name=\"%s\"\r\n", bridge_name);
	p += sprintf(p, "buildtype=\"test\"\r\n");
	p += sprintf(p, "macaddr=\"%s\"\r\n", lan_mac);

	// added field in spec. v1.0
	p += sprintf(p, "mgmt_ip=\"%s\"\r\n", nvram_safe_get("lan_ipaddr_stb"));
	p += sprintf(p, "mgmt_subnet=\"%s\"\r\n", nvram_safe_get("lan_netmask_stb"));
	p += sprintf(p, "dhcpclient=\"%s\"\r\n", nvram_safe_get("sta_dhcp"));
	p += sprintf(p, "ip=\"%s\"\r\n", is_def == 1 ? "" : lan_ip);
	p += sprintf(p, "subnet=\"%s\"\r\n", is_def == 1 ? "" : lan_mask);
	p += sprintf(p, "gateway=\"%s\"\r\n", is_def == 1 ? "" : gw);
	p += sprintf(p, "dns=\"%s\"\r\n",is_def == 1 ? "" : get_dns());

	int isLink = read_assoc_status(ssid, security, encryption, channel, mode, apmac, strength);
	if( isLink == ASSOC_STATUS_ERROR )  // error
		return 1;
	if( isLink == ASSOC_STATUS_OK && strcmp("00:00:00:00:00:00", apmac) == 0 ) // if the apmac is 00:00:00:00:00:00, it not associated.
		isLink = ASSOC_STATUS_NOT;
       
	if( isLink == ASSOC_STATUS_OK ) {
		p += sprintf(p, "ssid=\"%s\"\r\n", ssid);
		p += sprintf(p, "security=\"%s\"\r\n", security);
		p += sprintf(p, "encryption=\"%s\"\r\n", encryption);
		p += sprintf(p, "channel=\"%s\"\r\n", channel);
		p += sprintf(p, "mode=\"%s\"\r\n", mode);
		p += sprintf(p, "apmac=\"%s\"\r\n", apmac);
		p += sprintf(p, "strength=\"%s\"\r\n", strength);
	} else { // not assoc
		p += sprintf(p, "ssid=\"\"\r\n");
		p += sprintf(p, "security=\"\"\r\n");
		p += sprintf(p, "encryption=\"\"\r\n");
		p += sprintf(p, "channel=\"\"\r\n");
		p += sprintf(p, "mode=\"\"\r\n");
		p += sprintf(p, "apmac=\"\"\r\n");
		p += sprintf(p, "strength=\"\"\r\n");
	}

	if( isLink == ASSOC_STATUS_OK ) { // has joined to a alive ap
		p += sprintf(p, "state=\"AUTH\"\r\n");
		p += sprintf(p, "surveying=\"0\"\r\n");
		p += sprintf(p, "joining=\"0\"\r\n");
	} else if( isLink == ASSOC_STATUS_NOCONF ) {
		p += sprintf(p, "state=\"NOCONF\"\r\n");
		p += sprintf(p, "surveying=\"0\"\r\n");
		p += sprintf(p, "joining=\"0\"\r\n");
	} else { // not associate any ap
		p += sprintf(p, "state=\"JOIN\"\r\n");
		p += sprintf(p, "surveying=\"0\"\r\n");
		p += sprintf(p, "joining=\"1\"\r\n");
	}

	char ct[64] = {0};
	sprintf(ct, "Content-length: %d\r\n\r\n", strlen(buffer));

	print_header(req);
	req_write(req, ct);
	req_write(req, buffer);
	req_flush(req);

	return 0;
}

int stbadmin_capabilities(request *req, char *name[], char *value[], int sum)
{
	ASSERT_UPGRADE(1);

	FILE *fp = fopen("/tmp/boa_upgrade", "r"); // this file is created in src/read.c:191
	char buffer[128] = {0}, *p = buffer, cl[128] = {0};
	p += sprintf(p, "firmware=\"%d\"\r\n", m_gb_upgrade || fp ? 0 : 1);
	p += sprintf(p, "upgrade=\"1\"\r\n");
	p += sprintf(p, "dhcpclient=\"%s\"\r\n", nvram_safe_get("sta_dhcp"));

	sprintf(cl, "Content-length: %d\r\n\r\n", strlen(buffer));

	if( fp )
		fclose(fp);

	print_header(req);
	req_write(req, cl);
	req_write(req, buffer);
	req_flush(req);

	return 0;
}

int stbadmin_set(request *req, char *name[], char *value[], int sum)
{
	ASSERT_UPGRADE(0);

	int i = 0;
	char *ip = 0, *subnet = 0, *dhcpc = 0, *gw = 0, *dns = 0;
	char br0[8] = {0}, br0_1[8] = {0}, stb_ip[32] = {0}, stb_mask[32] = {0};
	char op[1024] = {0};
	static int back_dhcpc = 0;

	// backup the static setting when first time
	if( 0 == back_dhcpc ) {
		echo("First run the SET command, backup the static network settings.\n");
		nvram_set("back_lan_ipaddr", nvram_safe_get("lan_ipaddr"));
		nvram_set("back_lan_netmask", nvram_safe_get("lan_netmask"));
		nvram_set("back_lan_gateway", nvram_safe_get("lan_gateway"));
		nvram_set("back_sta_dhcp", nvram_safe_get("sta_dhcp"));
		nvram_set("back_", "1"); // no garbage
		back_dhcpc = 1;
	}

	strcpy(br0, nvram_safe_get("lan_ifname"));
	strcpy(br0_1, nvram_safe_get("lan_ifname_stb"));
	strcpy(stb_ip, nvram_safe_get("lan_ipaddr_stb"));
	strcpy(stb_mask, nvram_safe_get("lan_netmask_stb"));
	for(i = 0; i < sum ; i++) {
		if( !strcmp(name[i], "dhcpclient") && !strcmp(value[i], "1") ) {
			echo("User want enable the sta-DHCP, so we disable other params.\n");
			name[0] = name[i];
			value[0] = value[i];
			sum = 1;
			break;
		}
	}
	for(i = 0; i < sum ; i++) {
		if( !strcmp(name[i], "ip") ) {
			ip = value[i];
			if( !check_address_valid(value[i], 0) )
				return 1;
			nvram_set("lan_ipaddr", ip);
			sprintf(op, "ifconfig %s %s > /dev/console 2>&1", br0, ip);
			echo("%s\n", op);
			system(op);
		}

		if( !strcmp(name[i], "subnet") ) {
			subnet = value[i];
			if( !check_address_valid(value[i], 1) )
				return 1;
			nvram_set("lan_netmask", subnet);
			sprintf(op, "ifconfig %s netmask %s > /dev/console 2>&1", br0, subnet);
			echo("%s\n", op);
			system(op);
		}

		if( !strcmp(name[i], "dhcpclient") ) {
			dhcpc = value[i];
			if( *dhcpc != '0' && *dhcpc != '1' )
				return 1;
			nvram_set("sta_dhcp", dhcpc);
			sprintf(op, "lan restart > /dev/console 2>&1");
			echo("%s\n", op);
			system(op);
		}

		if( !strcmp(name[i], "gateway") ) {
			gw = value[i];
			if( !check_address_valid(value[i], 0) )
				return 1;
			nvram_set("lan_gateway", gw);
			sprintf(op, "route add default gw %s dev %s  > /dev/console 2>&1", gw, br0);
			echo("%s\n", op);
			system(op);
		}

		if( !strcmp(name[i], "dns") ) {
			dns = value[i];
			if( !check_address_valid(value[i], 0) )
				return 1;
			sprintf(op, "echo \"nameserver %s\" >> " RESOLV_CONF, dns);
			echo("%s\n", op);
			system(op);
		}
	}
	nvram_commit();

	// re-setup the Sky interface
	sprintf(op, "ifconfig %s %s netmask %s > /dev/console 2>&1", br0_1, stb_ip, stb_mask);
	echo("%s\n", op);
	system(op);

	print_header(req);
	req_write(req, "Content-length: 0\r\n\r\n");
	req_flush(req);

	return 0;
}

#define FILE_SCAN_RESULT "/tmp/tmp_scanresults_stbadmin"
int stbadmin_survey(request *req, char *name[], char *value[], int sum)
{
	ASSERT_UPGRADE(0);

	print_header(req);
	req_write(req, "Content-length: 0\r\n\r\n");
	req_flush(req);

	system("/usr/sbin/wl scan");
	sleep(2);
	system("/usr/sbin/wl scanresults > " FILE_SCAN_RESULT );

	return 0;
}

int stbadmin_networks(request *req, char *name[], char *value[], int sum)
{
	ASSERT_UPGRADE(0);

	if( sum != 1 || value[0] == NULL )
		return 1;
	int ret_count = 0, total = 0;
	if( sscanf(value[0], "%d", &ret_count) == 0 ) // no number exist
		return 1;
	if( ret_count < 0 || ret_count > 99 ) // wrong rang
		return 1;

	FILE *fp = fopen(FILE_SCAN_RESULT, "r");
	if( !fp ) {
		send_error_response(req, "500");
		return 0;
	}

	char buffer[1024 * 8] = {0}, *bp = buffer, tmp[1024 * 2] = {0}, *q;
	char ssid[32] = {0}, channel[4] = {0}, mode[4] = {0}, security[12] = {"open"}, encryption[8] = {"none"}, strength[4] = {0};
	int chan = 0, itmp = 0;
	int wpa = 0, rsn = 0;
	ListItem *head = NULL, *next = NULL;

	memset(buffer, 0, sizeof(buffer));
	memset(tmp, 0, sizeof(tmp));

	while( !feof(fp) ) {
		memset(tmp, 0, sizeof(tmp));
		q = fgets(tmp, sizeof(tmp), fp);
		if( !q )
			continue;

		// filter
		while( isblank(*q) )
			q++;
		q[strlen(q) - 1] = 0;  // the last \n character

		if( strncmp("SSID", q, 4) == 0 ) { // SSID
			q = strchr(q, ':');
			q += 3;
			q[strlen(q) - 1] = 0;
			strcpy(ssid, q);
			continue;
		}

		if( strncmp("Mode", q, 4) == 0 ) { // Mode, RSSI, Channel
			char *s_rssi = strstr(q, "RSSI") + 6;
			int d_rssi = 0;
			sscanf(s_rssi, "%d", &d_rssi);
			sprintf(strength, "%d", (itmp = (int)((d_rssi + 95)*100/85)) >= 100 ? 100 : itmp);

			q = strstr(q, "Channel") + strlen("Channel") + 2;
			strcpy(channel, q);
			sscanf(channel, "%d", &chan);

			continue;
		}

		if( strncmp("BSSID", q, 5) == 0 ) {
			q = strstr(q, "WEP");
			if( q ) {
				strcpy(security, "shared");
				strcpy(encryption, "wep");
			}

			continue;
		}

		if( strncmp("Supported Rates", q, 15) == 0 ) {
			// the rate format: [ 1(b) 2(b) 5.5(b) 6 9 11(b) 12 18 24 36 48 54 ]
			// b rate: 1, 2, 5.5, 11
			// g rate: 6, 9, 12, 18, 24, 36, 48, 54
			q += 15 + 2;

			int brate = 0, grate = 0, rate;
			char *pr = strtok(q, " ");
			while( pr ) {
				if( sscanf(pr, "%d", &rate) > 0 ) {
					switch( rate ) {
						case 1: case 2: case 5: case 11:
							brate = 1;
							break;
						case 6: case 9: case 12: case 18: case 24: case 36: case 48: case 54:
							grate = 1;
							break;
						default:
							echo("Wrong Rates");
							break;
					}
				}

				pr = strtok(NULL, " ");
			}
			if( brate == 1 && grate == 0 )
				strcpy(mode, "b");
			if( brate == 0 && grate == 1 )
				strcpy(mode, "g");
			if( brate && grate )
				strcpy(mode, "g");

			continue;
		}

		if( (wpa = (strncmp("WPA", q, 3) == 0)) || (rsn = (strncmp("RSN", q, 3) == 0)) ) {
			if( wpa ) { // have WPA or WPA2 block
				q += 3;
				if( *q == '2' ) { // wpa2
					strcpy(security, "wpa-psk2");
				} else { // wpa
					strcpy(security, "wpa-psk");
				}
			}
			if( rsn && security[0] != 'w' ) { // only has the "RSN" block, we will treat it as wpa2-psk
				strcpy(security, "wpa2-psk");
			}

			memset(tmp, 0, sizeof(tmp));
			q = fgets(tmp, sizeof(tmp), fp);
			if( (q = strstr(q, "TKIP")) )
				strcpy(encryption, "tkip");
			else
				strcpy(encryption, "aes");

			continue;
		}

		// a null line
		if( strlen(q) == 0 ) {
			//echo("ssid=\"%s\",channel=\"%s\",mode=\"%s\",security=\"%s\",encryption=\"%s\",strength=\"%s\"\n", 
			//	ssid, channel, mode, security, encryption, strength);
			if( !(networks_insert_item(&head, ssid, channel, mode, security, encryption, strength)) )
				total++;

			// clear
			memset(ssid, 0, sizeof(ssid));
			memset(channel, 0, sizeof(channel));
			memset(mode, 0, sizeof(mode));
			strcpy(security, "open");
			strcpy(encryption, "none");
			memset(strength, 0, sizeof(strength));

			if( total == 99 ) // max items
				break;
			continue;
		}
	}
	fclose(fp);

	// output to buffer
	itmp = (ret_count <= 0 ? total : ret_count);
	while( head ) {
		if( itmp-- > 0 ) {
			bp += sprintf(bp, "ssid=\"%s\",channel=\"%s\",mode=\"%s\",security=\"%s\",encryption=\"%s\",strength=\"%s\"\r\n",
				head->ssid, head->channel, head->mode, head->security, head->encryption, head->strength);
		}

		next = head->next;
		free(head);
		head = next;
	}
	bp += sprintf(bp, "\r\n");

	// send the response to client
	print_header(req);

	char cont[64] = {0};
	memset(tmp, 0, sizeof(tmp));
	sprintf(cont, "Content-length: %d\r\n\r\n", 
			sprintf(tmp, "nets_found=\"%d\"\r\nnets_returned=\"%d\"\r\n", total, (ret_count >= total || ret_count <= 0) ? total : ret_count) +
			strlen(buffer));
	req_write(req, cont);
	req_write(req, tmp);
	req_write(req, buffer);
	req_flush(req);

	return 0;
}

int stbadmin_join(request *req, char *name[], char *value[], int sum)
{
	ASSERT_UPGRADE(0);

	if( sum != 4 )
		return 1;

	int i = 0;
	char *p_ssid = 0, *p_security = 0, *p_encryption = 0, *p_key = 0;

	// check validation
	for(i = 0; i < 4 ; i++ ) {
		if( strcmp(name[i], "ssid") == 0 )
			p_ssid = value[i];
		if( strcmp(name[i], "security") == 0 )
			p_security = value[i];
		if( strcmp(name[i], "encryption") == 0 )
			p_encryption = value[i];
		if( strcmp(name[i], "key") == 0 )
			p_key = value[i];
	}
	if( !p_ssid || !p_security || !p_encryption || !p_key )
		return 1;
	if( unescape_uri(p_ssid, NULL) == 0 ) {
		echo("unescape_uri: ssid error.");
		return 1;
	}
	if( unescape_uri(p_key, NULL) == 0 ) {
		echo("unescape_uri: key error.");
		return 1;
	}
	if( valid_name(p_ssid, ARGV("0", "32")) == 0 ||
	    valid_choice(p_security, ARGV("open", "shared", "wpa-psk", "wpa2-psk")) == 0 ||
	    valid_choice(p_encryption, ARGV("none", "wep64", "wep128", "tkip", "aes")) == 0 )
		return 1;

	if( !strcmp(p_security, "shared") ) {
		if( (strcmp(p_encryption, "wep64") == 0 && valid_name(p_key, ARGV("10", "10")) == 0) || 
		    (strcmp(p_encryption, "wep128") == 0 && valid_name(p_key, ARGV("26", "26")) == 0) ) 
			return 1;
	}
	if( strncmp(p_security, "wpa", 3) == 0 && valid_name(p_key, ARGV("8", "63")) == 0 ) {
		return 1;
	}

	// setup, according to wg602v4-gui/cgi/security.cgi
	system("/sbin/nas-bin stop");

	nvram_set("wl_ssid", p_ssid);
	nvram_set("wl0_ssid", p_ssid);

	if( !strcmp(p_security, "open") && !strcmp(p_encryption, "none") ) {
		nvram_set("sta_sectype", "0");
		nvram_set("wl_sectype", "0");

		nvram_set("wl_wep", "disabled");
		nvram_set("wl0_wep", "disabled");
		nvram_set("wl_auth", "0");
		nvram_set("wl0_auth", "0"); // off
		nvram_set("wl_wpa_gtk_rekey", "0");
		nvram_set("wl0_wpa_gtk_rekey", "0");
		nvram_set("wl_auth_mode", "open");
		nvram_set("wl0_auth_mode", "open");
		nvram_set("wl_crypto", "tkip");
		nvram_set("wl0_crypto", "tkip");
		nvram_set("wl0_akm", "0");
	}

	if( (!strcmp(p_security, "shared") || !strcmp(p_security, "open")) && 
	    (!strcmp(p_encryption, "wep64") || !strcmp(p_encryption, "wep128")) ) {  // wep
		nvram_set("sta_sectype", "1");
		nvram_set("wl_sectype", "1");

		nvram_set("wl_wep", "enabled");
		nvram_set("wl0_wep", "enabled");
		nvram_set("wl_auth_mode", p_security);
		nvram_set("wl0_auth_mode", p_security);
		nvram_set("wl_auth", !strcmp(p_security, "shared") ? "3" : "2");
		nvram_set("wl0_auth", !strcmp(p_security, "shared") ? "1" : "0"); // wep
		nvram_set("wl0_akm", "");

		nvram_set("wl_key", "1");
		nvram_set("wl_key1", p_key);
		nvram_set("wl_key2", p_key);
		nvram_set("wl_key3", p_key);
		nvram_set("wl_key4", p_key);
		nvram_set("wl0_key", "1");
		nvram_set("wl0_key1", p_key);
		nvram_set("wl0_key2", p_key);
		nvram_set("wl0_key3", p_key);
		nvram_set("wl0_key4", p_key);

		if( !strcmp(p_encryption, "wep64") ) { // 64
			nvram_set("key_length", "64");
			nvram_set("wep_64_key1", p_key);
			nvram_set("wep_64_key2", p_key);
			nvram_set("wep_64_key3", p_key);
			nvram_set("wep_64_key4", p_key);
		} else { // 128
			nvram_set("key_length", "128");
			nvram_set("wep_128_key1", p_key);
			nvram_set("wep_128_key2", p_key);
			nvram_set("wep_128_key3", p_key);
			nvram_set("wep_128_key4", p_key);
		}
	}

	if( !strcmp(p_security, "wpa-psk") ) {
		nvram_set("sta_sectype", "2");
		nvram_set("wl_sectype", "2");

		char tmp[1024] = {0};
		sprintf(tmp, "/usr/sbin/wl set_pmk %s", p_key);
		system("/usr/sbin/wl wpa_auth psk");
		system(tmp);

		nvram_set("wl_wpa_psk", p_key);
		nvram_set("wl0_wpa_psk", p_key);
		nvram_set("wl_wpa1_psk", p_key);
		nvram_set("wl0_wpa1_psk", p_key);

		nvram_set("wl_wep", "disabled");
		nvram_set("wl0_wep", "disabled");
		nvram_set("wl_auth_mode", "psk");
		nvram_set("wl0_auth_mode", "psk");
		nvram_set("wl_auth", "0");
		nvram_set("wl0_auth", "0");
		nvram_set("wl0_akm", "psk");
		nvram_set("wl0_crypto", "tkip");
		nvram_set("wl0_wpa_gtk_rekey", "0");

		sprintf(tmp, "*,auto,tkip,psk,DNIRONGER,%s", p_key);
		nvram_set("wl0_wds0", tmp);

		system("/sbin/nas-bin start");
	}

	if( !strcmp(p_security, "wpa2-psk") ) {
		nvram_set("sta_sectype", "3");
		nvram_set("wl_sectype", "3");

		char tmp[1024] = {0};
		sprintf(tmp, "/usr/sbin/wl set_pmk %s", p_key);
		system("/usr/sbin/wl wpa_auth psk2");
		system(tmp);

		nvram_set("wl_wpa_psk", p_key);
		nvram_set("wl0_wpa_psk", p_key);
		nvram_set("wl_wpa2_psk", p_key);
		nvram_set("wl0_wpa2_psk", p_key);

		nvram_set("wl_wep", "disabled");
		nvram_set("wl0_wep", "disabled");
		nvram_set("wl_auth_mode", "psk2");
		nvram_set("wl0_auth_mode", "psk2");
		nvram_set("wl_auth", "0");
		nvram_set("wl0_auth", "0");
		nvram_set("wl0_akm", "psk2");
		nvram_set("wl0_crypto", "aes");
		nvram_set("wl0_wpa_gtk_rekey", "0");

		sprintf(tmp, "*,auto,aes,psk,DNIRONGER,%s", p_key);
		nvram_set("wl0_wds0", tmp);

		system("/sbin/nas-bin start");
	}

	nvram_commit();

	// send back the response
	print_header(req);
	req_write(req, "Content-length: 0\r\n\r\n");
	req_flush(req);

	// restart the wireless driver
	wireless_reboot(0);

	return 0;
}

int stbadmin_reboot(request *req, char *name[], char *value[], int sum)
{
	ASSERT_UPGRADE(0);

        print_header(req);
        req_write(req, "Content-length: 0\r\n\r\n");
        req_flush(req);

	sleep(2);
	system("/sbin/reboot");
	return 0;
}

int stbadmin_reset(request *req, char *name[], char *value[], int sum)
{
	ASSERT_UPGRADE(0);

	print_header(req);
	req_write(req, "Content-length: 0\r\n\r\n");
	req_flush(req);

	nvram_set("restore_defaults", "1");
	nvram_commit();

	sleep(2);
	system("/sbin/reboot");
	return 0;
}

int stbadmin_upgrade(request *req, char *name[], char *value[], int sum)
{
	ASSERT_UPGRADE(0);
	m_gb_upgrade = 1;

	if( sum != 1 || !value[0] || strlen(value[0]) == 0 ) {
		m_gb_upgrade = 0;
		return 1;
	}

	char *uri = value[0], tmp[1024] = {0};
	int type = 0;  // 1 -- http, 2 -- ftp

	// check validation
	memcpy(tmp, uri, 4);
	if( strcasecmp(tmp, "http") == 0 )
		type = 1;
	tmp[3] = 0;
	if( strcasecmp(tmp, "ftp") == 0 )
		type = 2;
	if( type == 0 ) {
		m_gb_upgrade = 0;
		return 1;  // not http or ftp, not supported
	}
	if( unescape_uri(uri, NULL) == 0 ) {
		m_gb_upgrade = 0;
		return 1;
	}

	// send back the response
	print_header(req);
	req_write(req, "Content-length: 0\r\n\r\n");
	req_flush(req);

	// upgrade
	unlink(UPGRADE_STATUS);
	sprintf(tmp, "/usr/sbin/fw-upgrade.sh %s > /dev/console 2>&1", uri);
	system(tmp);

	set_timer(5, spy_upgrade_handler);
	return 0;
}

typedef int (*method)(request *req, char *name[], char *value[], int sum);
typedef struct _operation {
	char opname[32];
	method func;
} operation;

operation list[] = {
	{"info",     stbadmin_info}, 
	{"capabilities",  stbadmin_capabilities}, 
	{"set",      stbadmin_set}, 
	{"survey",   stbadmin_survey},
	{"networks", stbadmin_networks},
	{"join",     stbadmin_join},
	{"reboot",   stbadmin_reboot},
	{"reset",    stbadmin_reset},
	{"upgrade",  stbadmin_upgrade},
	{"", NULL},
};

void stbadmin(request *req)
{
	char query[MAX_HEADER_LENGTH] = {0}, *param = 0, *name[8] = {0}, *value[8] = {0}; // suport max 8 name value pairs
	int i = 0, j = 0;

	strcpy(query, req->query_string);
	if( (param = strchr(query, '&')) ) { // have param
		*param++ = 0;

		for(i = 0; i < 8 && param ; i++ ) {
			name[i] = param;
			value[i] = strchr(param, '=') + 1;

			param = value[i] - 1;
			*param++ = 0;

			// next name=value pair
			param = strchr(param, '&');
			if( param )
				*param++ = 0;

			// git rid of the "", it has been escaped to %22
			if( *value[i] ) {
				if( value[i][0] == '%' && value[i][1] == '2' && value[i][2] == '2' ) { // %22
					value[i] = value[i] + 3;
					value[i][strlen(value[i]) - 3] = 0;
				} else if( value[i][0] == '"' ) {  // "
					value[i] = value[i] + 1;
					value[i][strlen(value[i]) - 1] = 0;
				}
			}
		}
	}

//	dump_uri(req, req->request_uri, req->query_string, name, value, "Dump URI");
//	return ;

	int retval = 0;
	for(j = 0; list[j].opname[0] ; j++ ) {
		if( strcasecmp(list[j].opname, query) == 0 && list[j].func ) {
			retval = (list[j].func)(req, name, value, i);
			break;
		}
	}

	if( retval || list[j].opname[0] == 0 ) {
//		dump_uri(req, req->request_uri, req->query_string, name, value, "Invalid argument.");  // just for debug and test
		send_r_not_found(req);
	}
}
